#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <stdio.h>  
#include <stdlib.h>   

using namespace std;

#define INTERPOLATION
//#define POISSON


// globale Variablen
#ifdef INTERPOLATION
const int polDegree = 2;
const int N = polDegree + 1;
const int M = 200;
#endif

#ifdef POISSON
const int N = 1000;
const int M = N + 2;
#define PWCONSTANT
//#define SINUS
#endif

const double EPS = 1e-15;
const int BW = 3;

//=========================================================

//! run pdflatex for given tex-file
bool runPdflatex ( const string& CommandFileName )  {
  string systemCommand = "pdflatex ";
  systemCommand += CommandFileName;
  cout << systemCommand << endl;
  bool failed = true;
  for ( unsigned short int i = 0; i < 2; ++i ) {
    failed = ( system ( systemCommand.c_str() ) != EXIT_SUCCESS );
    if ( failed )
      cerr << "runPdflatex: Calling pdflatex returned an error.\n";
  }
  return !failed;
}

// write 2D data
void write2D( const double x[M], const double y[M], const char* filename ) {
  ofstream out ( filename );
  for( int i = 0; i < M; i++ )
    out << x[i] << " " << y[i] << endl;
}

//
void sampleSinusData( double x[M], double y[M] ){
  double eps = 1.;
  double min = 0 - eps;
  double max = 2. * M_PI + eps;
  double h = 1. * (max - min) / (M-1);
  
  for( int i = 0; i < M; i++ ){
    x[i] = min + i * h;
    y[i] = sin( x[i] );
  }
}

//
void samplePolynomialData( double coeff[N], double x[M], double y[M] ) {
  double eps = 1.;
  double min = 0 - eps;
  double max = 2. * M_PI + eps;
  double h = 1. * (max - min) / (M-1);
  
  for( int i = 0; i < M; i++ ){
    x[i] = min + i * h;
    double potX = 1.;
    y[i] = 0;
    for( int j = 0; j < N; j++ ){
      y[i] += coeff[j] * potX;
      potX *= x[i];
    }
  }
}

//=========================================================
//=========================================================
//=========================================================


//
void computeLR( double A[N][N], int per[N] ) {

  // initialize permutation vector
  for ( int j = 0; j < N; j++ )
    per[j] = j;
  
  // n-1 Iterationsschritte 
  for ( int i = 0; i < N; i++ ){ 
    
    // find pivot element
    double max = A[i][i];
    int pivot = i;
    for( int j = i; j < N; j++  ){
      if( std::abs( A[j][i] ) > max ){
	max = std::abs( A[j][i] );
	pivot = j;
      }
    }
    
    //
    if( max < EPS )
      throw "ERROR in computeLR()! Matrix is singular!";
    
    // swap rows
    std::swap( per[i], per[pivot] );
    for ( int j = 0; j < N; j++ )
      std::swap( A[pivot][j], A[i][j] );
    
    // Zeilen der Restmatrix werden durchlaufen
    for ( int k = i+1; k < N; k++ ){
      // Berechnung von L
      A[k][i] /= A[i][i];
      // Berechnung von R 
      for ( int j = i+1; j < N; j++ )         
         A[k][j] -= A[k][i] * A[i][j];      
    }
    
  }

}

//
void GaussElimination( double A[N][N], double rhs[N], double sol[N] ){  
 
  // compute LR factorization
  int per[N];
  computeLR( A, per );   
  
  // forward
  double y[N];
  for( int i = 0; i < N; i++ ){
    double sum = 0.;
    for( int j = 0; j < i; j++ )
      sum += A[i][j] * y[j];
    y[i] = rhs[ per[i] ] - sum;
  }
    
  // backward
  for( int i = N-1; i != -1; i-- ){
    double sum = 0.;
    for( int j = i+1; j < N; j++ )
      sum += A[i][j] * sol[j];
    sol[i] = (y[i] - sum) / A[i][i];
  }
  
}


//=========================================================

//
void computeLRForBandmatrixWithoutPivot( double A[BW][N] ) {
  
  // band width should be odd
  if( BW%2 != 1 )
    throw "ERROR in computeLRForBandmatrix! Bandwidth should be odd!";
  int diagIdx = (BW-1) / 2;
  
  // n-1 Iterationsschritte 
  for ( int i = 0; i < N - 1; i++ ){        
    // Zeilen der Restmatrix werden durchlaufen
    for ( int k = diagIdx+1; k < BW; k++ ){
      // Berechnung von L
      A[k][i] /= A[diagIdx][i];
      // Berechnung von R        
      for ( int j = 0; j < diagIdx; j++ ){         
	if( !(i+j+1 < N) ) 
	  continue;
        A[k-j-1][i+j+1] -= A[k][i] * A[diagIdx-j-1][i+j+1];      
      }
    }    
  }

}

//
void GaussEliminationForBandmatrixWithoutPivot( double A[BW][N], double rhs[N], double sol[N] ){  
 
  // band width should be odd
  if( BW%2 != 1 )
    throw "ERROR in GaussEliminationForBandmatrixWithoutPivot! Bandwidth should be odd!";
  int diagIdx = (BW-1) / 2;
  
  // compute LR factorization
  computeLRForBandmatrixWithoutPivot( A );   
  
  // forward
  double y[N];
  for( int i = 0; i < N; i++ ){
    double sum = 0.;
    for( int j = max(0, diagIdx - i); j < diagIdx; j++ )
      sum += A[BW-j-1][i-diagIdx+j] * y[i-diagIdx+j];
    y[i] = rhs[ i ] - sum;
  }
  
  // backward
  for( int i = N-1; i != -1; i-- ){
    double sum = 0.;
    for( int j = 0; j < min(diagIdx, N-1-i); j++ ) 
      sum += A[diagIdx-j-1][i+j+1] * sol[i+j+1];
    sol[i] = (y[i] - sum) / A[diagIdx][i];
  }
  
}

//=========================================================

//
void buildSystemForSinusInterpolation( double A[N][N], double rhs[N] ) {
   double max = 2 * M_PI;
   for( int i = 0; i < N; i++ ){
    A[i][0] = 1.;
    for( int j = 1; j < N; j++ )
      A[i][j] = A[i][j-1] * i * max / (N-1);
    rhs[i] = sin( i * max / (N-1) );
   }
}

//
void buildSystemForPoissonProblem( double A[N][N], double rhs[N] ) {
   for( int i = 0; i < N; i++ ){
     
#ifdef PWCONSTANT     
    // rhs
    if ( (i > N/4) && (i < 3*N/4) )
      rhs[i] = 1.;
    else
      rhs[i] = 0.;
#endif
     
#ifdef SINUS    
    // rhs
    double x = M_PI * (i+1) / (N+1);
    rhs[i] = sin(x) / x;
#endif
    
    // Diagonale und Nebendiagonale
    A[i][i] = 2. * N * N;
    if( i > 0 )
      A[i][i-1] = -1. * N * N;
    if( i < N-1 )
      A[i][i+1] = -1. * N * N;
    
    // Nulleintraege
    for( int j = i+2; j < N; j++ )
      A[i][j] = 0.;
    for( int j = 0; j < i-1; j++ )
      A[i][j] = 0.;
   }
}

//
void buildSystemForPoissonProblemBM( double A[BW][N], double rhs[N] ) {
   for( int i = 0; i < N; i++ ){
     
#ifdef PWCONSTANT     
    // rhs
    if ( (i > N/4) && (i < 3*N/4) )
      rhs[i] = 1.;
    else
      rhs[i] = 0.;
#endif
     
#ifdef SINUS    
    // rhs
    double x = M_PI * (i+1) / (N+1);
    rhs[i] = sin(x) / x;
#endif
    
    // band matrix
    for( int i = 0; i < N; i++ ){
      A[0][i] = -1. * N * N;
      A[1][i] = 2 * N * N;
      A[2][i] = -1 * N * N;
    }

   }
}

//=========================================================
//=========================================================

// Hauptprogramm
int main () {
    
  try{      
    
  static double A[N][N];
  static double rhs[N];
  static double sol[N];    

 clock_t begin, end;
  
#ifdef INTERPOLATION  
  buildSystemForSinusInterpolation( A, rhs );  
#endif
  
#ifdef POISSON  
  buildSystemForPoissonProblem( A, rhs );
#endif
  
  begin = clock();   
  GaussElimination( A, rhs, sol );  
  end = clock();
  
  cerr << "Time for greedy Gauss = " << double(end - begin) / CLOCKS_PER_SEC << endl;

    
  static double x[M];
  static double y[M];
  static double z[M];
  static double g[M];
  
  double a = 0;
  double b = 0;
  
  
#ifdef INTERPOLATION  
  sampleSinusData( x, y );
  samplePolynomialData( sol, x, z );
#endif
  
#ifdef POISSON  
  for( int j = 0; j < M; j++ )
    x[j] = 1. * j / (M-1); 
  
  y[0] = rhs[0] > 0 ? 1 : 0;
  y[N+1] = 0;
  for( int j = 0; j < N; j++ )
    y[j+1] = rhs[j];
  
  z[0] = a;
  z[N+1] = b;
  for( int j = 0; j < N; j++ )
    z[j+1] = sol[j] + (b-a) * x[j+1] + a;  
#endif
    
  write2D( x, y, "data1.dat" );
  write2D( x, z, "data2.dat" );
  //runPdflatex ( "plotResults.tex" );
  
#ifdef POISSON  
  //
  static double B[BW][N];
  buildSystemForPoissonProblemBM( B, rhs );    
  
  begin = clock();
  GaussEliminationForBandmatrixWithoutPivot( B, rhs, sol );
  end = clock();
  
  cerr << "Time for Gauss using band structure = " << double(end - begin) / CLOCKS_PER_SEC << endl;
#endif
    
  }
  catch ( const char* exc ){
    cerr << "Exception caught: " << exc << endl;
  }

  return 0;
}